package org.basex.query.func.update;

import org.basex.query.func.map.*;

/**
 * Function implementation.
 *
 * @author BaseX Team 2005-22, BSD License
 * @author Christian Gruen
 */
public class UpdateMapForEach extends MapForEach {
}
