package org.basex.query.func.db;

import static org.basex.query.QueryError.*;

import org.basex.data.*;
import org.basex.io.*;
import org.basex.query.*;
import org.basex.query.value.item.*;
import org.basex.util.*;

/**
 * Function implementation.
 *
 * @author BaseX Team 2005-22, BSD License
 * @author Christian Gruen
 */
public final class DbRetrieve extends DbAccess {
  @Override
  public B64Lazy item(final QueryContext qc, final InputInfo ii) throws QueryException {
    final Data data = toData(qc);
    final String path = toDbPath(1, qc);
    if(data.inMemory()) throw DB_MAINMEM_X.get(info, data.meta.name);

    final IOFile bin = data.meta.binary(path);
    if(!bin.exists() || bin.isDir()) throw WHICHRES_X.get(info, path);
    return new B64Lazy(bin, IOERR_X);
  }
}
