package org.basex.query.func;

import org.basex.*;
import org.junit.jupiter.api.*;

/**
 * This class tests the functions of the SQL Module.
 * Due to the lack of a default JDBC driver, this class contains only tests signatures.
 *
 * @author BaseX Team 2005-22, BSD License
 * @author Christian Gruen
 */
public final class SqlModuleTest extends SandboxTest {
  /** Test method. */
  @Test public void close() {
  }

  /** Test method. */
  @Test public void commit() {
  }

  /** Test method. */
  @Test public void connect() {
  }

  /** Test method. */
  @Test public void execute() {
  }

  /** Test method. */
  @Test public void executePrepared() {
  }

  /** Test method. */
  @Test public void init() {
  }

  /** Test method. */
  @Test public void prepare() {
  }

  /** Test method. */
  @Test public void rollback() {
  }
}
