
/**
 * This module contains {@code JavaCC}-generated classes for parsing XSD regular
 * expressions. All files are under BSD license, although proper headers can't always be
 * guaranteed.
 *
 * @author BaseX Team 2005-22, BSD License
 * @author Leo Woerteler
 */
package org.basex.query.util.regex.parse;