package org.basex.tests.w3c;

/**
 * Strings.
 *
 * @author BaseX Team 2005-22, BSD License
 * @author Christian Gruen
 */
interface QT3Constants {
  /** QT3TS String. */
  String CATALOG = "catalog.xml";

  /** QT3TS String. */
  String NNAME = "name";
  /** QT3TS String. */
  String SOURCE = "source";
  /** QT3TS String. */
  String RESOURCE = "resource";
  /** QT3TS String. */
  String PARAM = "param";
  /** QT3TS String. */
  String NAMESPACE = "namespace";
  /** QT3TS String. */
  String SCHEMA = "schema";
  /** QT3TS String. */
  String COLLATION = "collation";
  /** QT3TS String. */
  String STATIC_BASE_URI = "static-base-uri";
  /** QT3TS String. */
  String DECLARED = "declared";

  /** QT3TS String. */
  String FILE = "file";
  /** QT3TS String. */
  String ROLE = "role";
  /** QT3TS String. */
  String CODE = "code";
  /** QT3TS String. */
  String URI = "uri";
  /** QT3TS String. */
  String SELECT = "select";
  /** QT3TS String. */
  String PREFIX = "prefix";
  /** QT3TS String. */
  String ENCODING = "encoding";
}
