/* Generated By:JavaCC: Do not edit this line. TokenManager.java Version 5.0 */
/* JavaCCOptions:SUPPORT_CLASS_VISIBILITY_PUBLIC=true */
package org.basex.query.util.regex.parse;

/**
 * An implementation for this interface is generated by
 * JavaCCParser.  The user is free to use any implementation
 * of their choice.
 */
public interface TokenManager {
  /**
   * This gets the next token from the input stream.
   * A token of kind 0 (&lt;EOF&gt;) should be returned on EOF.
   * @return the next token
   */
  Token getNextToken();

}
/* JavaCC - OriginalChecksum=06afe96651fae510c1b250bfb8f28360 (do not edit this line) */
