/* Generated By:JavaCC: Do not edit this line. RegExParserConstants.java */
package org.basex.query.util.regex.parse;


/**
 * Token literal values and constants.
 * Generated by org.javacc.parser.OtherFilesGen#start()
 */
public interface RegExParserConstants {

  /** End of File. */
  int EOF = 0;
  /** RegularExpression Id. */
  int OR = 1;
  /** RegularExpression Id. */
  int Q_MARK = 2;
  /** RegularExpression Id. */
  int STAR = 3;
  /** RegularExpression Id. */
  int PLUS = 4;
  /** RegularExpression Id. */
  int QUANT_OPEN = 5;
  /** RegularExpression Id. */
  int QUANT_CLOSE = 6;
  /** RegularExpression Id. */
  int NUMBER = 7;
  /** RegularExpression Id. */
  int COMMA = 8;
  /** RegularExpression Id. */
  int NPAR_OPEN = 9;
  /** RegularExpression Id. */
  int PAR_CLOSE = 10;
  /** RegularExpression Id. */
  int PAR_OPEN = 11;
  /** RegularExpression Id. */
  int CHAR = 12;
  /** RegularExpression Id. */
  int DIGIT = 13;
  /** RegularExpression Id. */
  int BACK_REF = 14;
  /** RegularExpression Id. */
  int WILDCARD = 15;
  /** RegularExpression Id. */
  int LINE_START = 16;
  /** RegularExpression Id. */
  int LINE_END = 17;
  /** RegularExpression Id. */
  int SINGLE_ESC = 18;
  /** RegularExpression Id. */
  int MULTI_ESC = 19;
  /** RegularExpression Id. */
  int CAT_ESC = 20;
  /** RegularExpression Id. */
  int BR_OPEN = 21;
  /** RegularExpression Id. */
  int NEG = 22;
  /** RegularExpression Id. */
  int TO = 23;
  /** RegularExpression Id. */
  int BR_CLOSE = 24;

  /** Literal token values. */
  String[] tokenImage = {
    "<EOF>",
    "<OR>",
    "<Q_MARK>",
    "<STAR>",
    "<PLUS>",
    "<QUANT_OPEN>",
    "<QUANT_CLOSE>",
    "<NUMBER>",
    "<COMMA>",
    "<NPAR_OPEN>",
    "<PAR_CLOSE>",
    "<PAR_OPEN>",
    "<CHAR>",
    "<DIGIT>",
    "<BACK_REF>",
    "<WILDCARD>",
    "<LINE_START>",
    "<LINE_END>",
    "<SINGLE_ESC>",
    "<MULTI_ESC>",
    "<CAT_ESC>",
    "<BR_OPEN>",
    "<NEG>",
    "<TO>",
    "<BR_CLOSE>",
  };

}
